/* this file holds the VNC/NetWinder specific irq information,
 separate from irq.c, to prevent polluting irq.c for other
 platforms */

#define assert(x) { if (!(x)) printk("assert failure! line %d\n",__LINE__); }

/*
 * Support for interrupt services
 *
 */

/* FootBridge translation tables
       Converts from our IRQ numbers into FootBridge masks
       !! These hex patterns are now #defined in irq.h !!
*/
int fb_irq_mask[16] = {
        0,              // 0 not used
        2,              // IRQ_SWI      1
        0x0000000C,     // IRQ_UART_DEBUG
        0,              // 3 not used
        0x00000010,     // IRQ_TIMER0
        0x00000020,     // IRQ_TIMER1
        0x00000040,     // IRQ_TIMER2
        0x00000080,     // IRQ_WATCHDOG
        0x00000100,     // IRQ_ETHER10
        0x00000200,     // IRQ_ETHER100
        0x00000400,     // IRQ_VIDCOMP
        0x00000800,     // IRQ_EXTERN_IRQ
        0x00030000,     // IRQ_DMA1
        0,              // 13 not used
        0,              // 14 not used
        0xF0000000      // IRQ_PCI_ERR
};

/* Conversion from our IRQ numbers into PIC mask is easy, because
       there are no gaps. Given an IRQ (n) that is 16<=n<=31,
       if (n<24) lo_mask = 1 << (n-16)
       else hi_mask = 1 << (n-24)
*/

/* In the following code, the irq number is split into three "zones".
       The first zone deals with interrupts sourced by the FootBridge.
       The second and third zones deal with irq's from the PIC (intel-type)
       controller */
//static unsigned long g_irq_mask;

/**
       @param

       @return
*/

/**
       Mask off the hardware interrupt

       @param irq the interrupt number
*/
void mask_irq(unsigned int irq) {
        unsigned char t;

        // we should never turn of the chaining irq's
        assert (irq!=IRQ_EXTERN_IRQ);
        assert (irq!=IRQ_PIC_HI);

//     g_irq_mask |= (1 << irq);

        if (irq >= 16)
                *IRQ_CLEAR = fb_irq_mask[irq-16];
        else if (irq < 8) {
                t = inb(PIC_MASK_LO);
                t |= (1 << (irq));
                outb(t, PIC_MASK_LO);
        } else {
                t = inb(PIC_MASK_HI);
                t |= (1 << (irq - 8));
                outb(t, PIC_MASK_HI);
        }
}

/**
       Unmask the hardware interrupt; i.e. allow these interrupts.
       @param irq the interrupt number
*/
void unmask_irq(unsigned int irq) {
        unsigned char t;

//     g_irq_mask &= ~(1 << irq);

        if (irq >= 16)
                *IRQ_SET = fb_irq_mask[irq-16];
        else if (irq < 8) {
                t = inb(PIC_MASK_LO);
                t &= ~(1 << (irq));
                outb(t, PIC_MASK_LO);
        } else {
                t = inb(PIC_MASK_HI);
                t &= ~(1 << (irq - 8));
                outb(t, PIC_MASK_HI);
        }
}

/**
       Initialize all the interrupts as OFF.
*/
void irq_init_irq(void) {
        unsigned long flags;

        save_flags (flags);
        cli ();

        // initialize the PIC hardware
        outb(0x11,PIC_LO);      // start init
        outb(0x00,PIC_MASK_LO); // vectors 0-7
        outb(0x04,PIC_MASK_LO); // chain down comes through 4
        outb(1,PIC_MASK_LO);    // done init
        outb(0x11,PIC_HI);
        outb(0x08,PIC_MASK_HI); // vectors 8-15
        outb(0x02,PIC_MASK_HI); // chain goes out on 2
        outb(1,PIC_MASK_HI);    // done init

        /* clear all interrupt enables
                except the two chaining ones */
        *IRQ_CLEAR = ~IRQ_EXTERN_IRQ_MASK;
        *IRQ_SET = IRQ_EXTERN_IRQ_MASK;
        outb(0xFF - 4, PIC_MASK_LO);
        outb(0xFF, PIC_MASK_HI);
        restore_flags (flags);

//     g_irq_mask = 0xFFFFFFFF - (1<<IRQ_EXTERN_IRQ) - (1<<IRQ_PIC_HI);
}


/**
       From what we know of the hardware, build a 32 bit mask
       showing the current enabled irq's

       @return a bit mask with each 1 showing a currently enabled irq
*/
unsigned long get_enabled_irqs(void) {
        unsigned long fb_raw, u, ret_val = 0;
        unsigned char pic_lo, pic_hi;

        fb_raw = *IRQ_SHOW ;
        pic_lo = ~inb(PIC_MASK_LO);
        pic_hi = ~inb(PIC_MASK_HI);

        // now, run through the fb ones and build
        // a mask of logical bits
        for (u=0; u<16; u++) {
                if (fb_raw & fb_irq_mask[u])
                        ret_val |= (1<<(u+16));
        }

        // the PIC irqs are one-for-one with the logicals
        ret_val |= ((pic_lo));
        ret_val |= ((pic_hi) << 8);

        return ret_val;
}
/*
 * This builds up the IRQ handler stubs using some ugly macros in irq.h
 *
 * These macros create the low-level assembly IRQ routines that do all
 * the operations that are needed to keep the AT interrupt-controller
 * happy. They are also written to be fast - and to disable interrupts
 * as little as humanly possible.
 *
 * NOTE! These macros expand to three different handlers for each line: one
 * complete handler that does all the fancy stuff (including signal handling),
 * and one fast handler that is meant for simple IRQ's that want to be
 * atomic. The specific handler is chosen depending on the SA_INTERRUPT
 * flag when installing a handler. Finally, one "bad interrupt" handler, that
 * is used when no handler is present.
 *
 * The timer interrupt is handled specially to insure that the jiffies
 * variable is updated at all times.  Specifically, the timer interrupt is
 * just like the complete handlers except that it is invoked with interrupts
 * disabled and should never re-enable them.  If other interrupts were
 * allowed to be processed while the timer interrupt is active, then the
 * other interrupts would have to avoid using the jiffies variable for delay
 * and interval timing operations to avoid hanging the system.
 */


/*
 * Pointers to the low-level handlers: first the general ones, then the
 * fast ones, then the bad ones.
 */
static void (* const interrupt[NR_IRQS])(void) = {
    IRQ_INTERRUPT( 0), IRQ_INTERRUPT( 1), IRQ_INTERRUPT( 2), IRQ_INTERRUPT( 3),
    IRQ_INTERRUPT( 4), IRQ_INTERRUPT( 5), IRQ_INTERRUPT( 6), IRQ_INTERRUPT( 7),
    IRQ_INTERRUPT( 8), IRQ_INTERRUPT( 9), IRQ_INTERRUPT(10), IRQ_INTERRUPT(11),
    IRQ_INTERRUPT(12), IRQ_INTERRUPT(13), IRQ_INTERRUPT(14), IRQ_INTERRUPT(15),
    IRQ_INTERRUPT(16), IRQ_INTERRUPT(17), IRQ_INTERRUPT(18), IRQ_INTERRUPT(19),
    IRQ_INTERRUPT(20), IRQ_INTERRUPT(21), IRQ_INTERRUPT(22), IRQ_INTERRUPT(23),
    IRQ_INTERRUPT(24), IRQ_INTERRUPT(25), IRQ_INTERRUPT(26), IRQ_INTERRUPT(27),
    IRQ_INTERRUPT(28), IRQ_INTERRUPT(29), IRQ_INTERRUPT(30), IRQ_INTERRUPT(31)
};

static void (* const fast_interrupt[NR_IRQS])(void) = {
    FAST_INTERRUPT( 0), FAST_INTERRUPT( 1), FAST_INTERRUPT( 2), FAST_INTERRUPT( 3),
    FAST_INTERRUPT( 4), FAST_INTERRUPT( 5), FAST_INTERRUPT( 6), FAST_INTERRUPT( 7),
    FAST_INTERRUPT( 8), FAST_INTERRUPT( 9), FAST_INTERRUPT(10), FAST_INTERRUPT(11),
    FAST_INTERRUPT(12), FAST_INTERRUPT(13), FAST_INTERRUPT(14), FAST_INTERRUPT(15),
    FAST_INTERRUPT(16), FAST_INTERRUPT(17), FAST_INTERRUPT(18), FAST_INTERRUPT(19),
    FAST_INTERRUPT(20), FAST_INTERRUPT(21), FAST_INTERRUPT(22), FAST_INTERRUPT(23),
    FAST_INTERRUPT(24), FAST_INTERRUPT(25), FAST_INTERRUPT(26), FAST_INTERRUPT(27),
    FAST_INTERRUPT(28), FAST_INTERRUPT(29), FAST_INTERRUPT(30), FAST_INTERRUPT(31)
};

static void (* const bad_interrupt[NR_IRQS])(void) = {
    BAD_INTERRUPT( 0), BAD_INTERRUPT( 1), BAD_INTERRUPT( 2), BAD_INTERRUPT( 3),
    BAD_INTERRUPT( 4), BAD_INTERRUPT( 5), BAD_INTERRUPT( 6), BAD_INTERRUPT( 7),
    BAD_INTERRUPT( 8), BAD_INTERRUPT( 9), BAD_INTERRUPT(10), BAD_INTERRUPT(11),
    BAD_INTERRUPT(12), BAD_INTERRUPT(13), BAD_INTERRUPT(14), BAD_INTERRUPT(15),
    BAD_INTERRUPT(16), BAD_INTERRUPT(17), BAD_INTERRUPT(18), BAD_INTERRUPT(19),
    BAD_INTERRUPT(20), BAD_INTERRUPT(21), BAD_INTERRUPT(22), BAD_INTERRUPT(23),
    BAD_INTERRUPT(24), BAD_INTERRUPT(25), BAD_INTERRUPT(26), BAD_INTERRUPT(27),
    BAD_INTERRUPT(28), BAD_INTERRUPT(29), BAD_INTERRUPT(30), BAD_INTERRUPT(31)
};

static void (* const probe_interrupt[NR_IRQS])(void) = {
    PROBE_INTERRUPT( 0), PROBE_INTERRUPT( 1), PROBE_INTERRUPT( 2), PROBE_INTERRUPT( 3),
    PROBE_INTERRUPT( 4), PROBE_INTERRUPT( 5), PROBE_INTERRUPT( 6), PROBE_INTERRUPT( 7),
    PROBE_INTERRUPT( 8), PROBE_INTERRUPT( 9), PROBE_INTERRUPT(10), PROBE_INTERRUPT(11),
    PROBE_INTERRUPT(12), PROBE_INTERRUPT(13), PROBE_INTERRUPT(14), PROBE_INTERRUPT(15),
    PROBE_INTERRUPT(16), PROBE_INTERRUPT(17), PROBE_INTERRUPT(18), PROBE_INTERRUPT(19),
    PROBE_INTERRUPT(20), PROBE_INTERRUPT(21), PROBE_INTERRUPT(22), PROBE_INTERRUPT(23),
    PROBE_INTERRUPT(24), PROBE_INTERRUPT(25), PROBE_INTERRUPT(26), PROBE_INTERRUPT(27),
    PROBE_INTERRUPT(28), PROBE_INTERRUPT(29), PROBE_INTERRUPT(30), PROBE_INTERRUPT(31)
};

/*
 * Initial irq handlers.
 */

static void no_action (int irq, void *dev_id, struct pt_regs *regs)
{
}

struct irqaction *irq_action[NR_IRQS];

unsigned long validirqs[0] = {
        (1 << IRQ_HARDDISK)     | (1 << IRQ_SOUND)      | (1 << IRQ_VGA)
        | (1 << IRQ_POWERLOW)   | (1 << IRQ_RTC_ALARM)  | (1 << IRQ_PRINTER)
        | (1 << IRQ_UART_IR)    | (1 << IRQ_MOUSE)      | (1 << IRQ_UART)
        | (1 << IRQ_UART2)      | (1 << IRQ_KEYBOARD)   | (1 << IRQ_BUTTON)
        | (1 << IRQ_DMA1)       | (1 << IRQ_VIDCOMP)    | (1 << IRQ_ETHER100)
        | (1 << IRQ_ETHER10)    | (1 << IRQ_UART_DEBUG)
        /* we probably do not want people playing around with these during
         * a probe; so take them out of the --valid-- list
        | (1 << IRQ_SWI)        | (1 << IRQ_PCI_ERR)
        | (1 << IRQ_WATCHDOG)   | (1 << IRQ_TIMER2)
        | (1 << IRQ_TIMER1)     | (1 << IRQ_TIMER0)
        | (1 << IRQ_TIMER4)
        */
        /* old old old old old old old
        //0x5bfa97f6            // lines up with what we consider valid irq's
        // prevents people from playing with the chained interrupts 0x00040800
        // !!NOTE TO SELF: irq 16 (timer 4) is disabled because it's too annoying (during probe)
        // and we don't have a use for it yet
        */
};

void (*irqjump[NR_IRQS])(void) = {
    PROBE_INTERRUPT( 0), PROBE_INTERRUPT( 1), PROBE_INTERRUPT( 2), PROBE_INTERRUPT( 3),
    PROBE_INTERRUPT( 4), PROBE_INTERRUPT( 5), PROBE_INTERRUPT( 6), PROBE_INTERRUPT( 7),
    PROBE_INTERRUPT( 8), PROBE_INTERRUPT( 9), PROBE_INTERRUPT(10), PROBE_INTERRUPT(11),
    PROBE_INTERRUPT(12), PROBE_INTERRUPT(13), PROBE_INTERRUPT(14), PROBE_INTERRUPT(15),
    PROBE_INTERRUPT(16), PROBE_INTERRUPT(17), PROBE_INTERRUPT(18), PROBE_INTERRUPT(19),
    PROBE_INTERRUPT(20), PROBE_INTERRUPT(21), PROBE_INTERRUPT(22), PROBE_INTERRUPT(23),
    PROBE_INTERRUPT(24), PROBE_INTERRUPT(25), PROBE_INTERRUPT(26), PROBE_INTERRUPT(27),
    PROBE_INTERRUPT(28), PROBE_INTERRUPT(29), PROBE_INTERRUPT(30), PROBE_INTERRUPT(31)
};
