/*
 *  linux/arch/arm/kernel/time.c
 *
 *  Copyright (C) 1991, 1992, 1995  Linus Torvalds
 *  Modifications for ARM (C) 1994, 1995, 1996,1997 Russell King
 *
 * This file contains the ARM-specific time handling details:
 * reading the RTC at bootup, etc...
 *
 * 1994-07-02  Alan Modra
 *             fixed set_rtc_mmss, fixed time.year for >= 2000, new mktime
 * 1997-09-10  Updated NTP code according to technical memorandum Jan '96
 *             "A Kernel Model for Precision Timekeeping" by Dave Mills
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/timex.h>
#include <asm/hardware.h>

#ifdef CONFIG_ARCH_VNC
#include <linux/mc146818rtc.h>
unsigned long get_cmos_time(void);
#endif


#ifndef BCD_TO_BIN
#define BCD_TO_BIN(val) ((val)=((val)&15) + ((val)>>4)*10)
#endif

#ifndef BIN_TO_BCD
#define BIN_TO_BCD(val) ((val)=(((val)/10)<<4) + (val)%10)
#endif

/* Converts Gregorian date to seconds since 1970-01-01 00:00:00.
 * Assumes input in normal date format, i.e. 1980-12-31 23:59:59
 * => year=1980, mon=12, day=31, hour=23, min=59, sec=59.
 *
 * [For the Julian calendar (which was used in Russia before 1917,
 * Britain & colonies before 1752, anywhere else before 1582,
 * and is still in use by some communities) leave out the
 * -year/100+year/400 terms, and add 10.]
 *
 * This algorithm was first published by Gauss (I think).
 *
 * WARNING: this function will overflow on 2106-02-07 06:28:16 on
 * machines were long is 32-bit! (However, as time_t is signed, we
 * will already get problems at other places on 2038-01-19 03:14:08)
 */
static inline unsigned long mktime(unsigned int year, unsigned int mon,
	unsigned int day, unsigned int hour,
	unsigned int min, unsigned int sec)
{
	if (0 >= (int) (mon -= 2)) {	/* 1..12 -> 11,12,1..10 */
		mon += 12;	/* Puts Feb last since it has leap day */
		year -= 1;
	}
	return (((
		    (unsigned long)(year/4 - year/100 + year/400 + 367*mon/12 + day) +
		      year*365 - 719499
		    )*24 + hour /* now have hours */
		   )*60 + min /* now have minutes */
		  )*60 + sec; /* finally seconds */
}

#include <asm/arch/time.h>

static unsigned long do_gettimeoffset(void)
{
	return gettimeoffset ();
}

void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;

	save_flags_cli (flags);
	*tv = xtime;
	tv->tv_usec += do_gettimeoffset();
	if (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
	restore_flags(flags);
}

void do_settimeofday(struct timeval *tv)
{
	cli ();
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= do_gettimeoffset();

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_state = TIME_ERROR;	/* p. 24, (a) */
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	sti ();
}

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick.
 */
static void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	if (reset_timer ())
            do_timer(regs);
        
#ifndef CONFIG_ARCH_VNC
        update_rtc ();
#endif
}

static struct irqaction irqtimer0 = { timer_interrupt, 0, 0, "timer", NULL, NULL};

extern int setup_arm_irq(int, struct irqaction *);

void time_init(void)
{
    xtime.tv_usec = 0;
#ifdef CONFIG_ARCH_VNC
    setup_timer();
    xtime.tv_sec = get_cmos_time();
    setup_arm_irq(IRQ_TIMER4, &irqtimer0);
#else
    xtime.tv_sec = setup_timer();
    setup_arm_irq(IRQ_TIMER0, &irqtimer0);
#endif
}

unsigned long get_cmos_time(void)
{
         unsigned int year, mon, day, hour, min, sec;
         int i;

        // check to see if the RTC makes sense.....
        if ((CMOS_READ(RTC_VALID) & RTC_VRT) == 0)
                return mktime(1970,1,1,0,0,0);

         /* The Linux interpretation of the CMOS clock register contents:
          * When the Update-In-Progress (UIP) flag goes from 1 to 0, the
          * RTC registers show the second which has precisely just started.
          * Let's hope other operating systems interpret the RTC the same way.
          */
         /* read RTC exactly on falling edge of update flag */
         for (i = 0 ; i < 1000000 ; i++) /* may take up to 1 second... */
                 if (CMOS_READ(RTC_FREQ_SELECT) & RTC_UIP)
                         break;
         for (i = 0 ; i < 1000000 ; i++) /* must try at least 2.228 ms */
                 if (!(CMOS_READ(RTC_FREQ_SELECT) & RTC_UIP))
                         break;
         do { /* Isn't this overkill ? UIP above should guarantee consistency */
                 sec = CMOS_READ(RTC_SECONDS);
                 min = CMOS_READ(RTC_MINUTES);
                 hour = CMOS_READ(RTC_HOURS);
                 day = CMOS_READ(RTC_DAY_OF_MONTH);
                 mon = CMOS_READ(RTC_MONTH);
                 year = CMOS_READ(RTC_YEAR);
         } while (sec != CMOS_READ(RTC_SECONDS));
         if (!(CMOS_READ(RTC_CONTROL) & RTC_DM_BINARY) || RTC_ALWAYS_BCD)
           {
             BCD_TO_BIN(sec);
             BCD_TO_BIN(min);
             BCD_TO_BIN(hour);
             BCD_TO_BIN(day);
             BCD_TO_BIN(mon);
             BCD_TO_BIN(year);
           }
         if ((year += 1900) < 1970)
                 year += 100;
         return mktime(year, mon, day, hour, min, sec);
}
